# R FUNCTIONS for continuous projections onto LPCs, SEE BELOW FOR AN EXAMPLE
# Ludger Evers/ Jochen Einbeck January-March 2008

# Connect LPC points by a cubic spline; contrary to the previous functions it just returns for each branch and coordinate a function that allows for evaluating the different splines, and most importantly, their derivatives (needed for computing the curve lengths)

# only available for unscaled version, i.e. scaled=FALSE in lpc.

lpc.splinefun <- function(lpcobject){
    nbranch <- max(lpcobject$Parameterization[,2])
    result <- list()
    for (r in 0:nbranch) {
      #print(range(lpcobject$Parameterization[lpcobject$Parameterization[,2]==r,1]))
      element <- list(branch=r, range=range(lpcobject$Parameterization[lpcobject$Parameterization[,2]==r,1]), splinefun=list())      
      for (j in 1:ncol(lpcobject$LPC)) 
        element$splinefun[[j]] <- splinefun(lpcobject$Parameterization[lpcobject$Parameterization[,2]==r,1], lpcobject$LPC[lpcobject$Parameterization[,2]==r,j])
      result[[r+1]] <- element
    }
    result
  }



# Evaluates the splines of the indicated branch at the original projection index or.pi

lpc.spline.eval <- function(lpcsl,or.pi,branch=0){
                          
    if (length(branch)==1)
       branch <- rep(branch,length(branch))
    result <- matrix(nrow=length(or.pi), ncol=length(lpcsl[[1]]$splinefun))
    branches <- unique(branch)
    for (cur.branch in branches) {
      cur.or.pi <- or.pi[branch==cur.branch]
      for (j in 1:length(lpcsl[[cur.branch+1]]$splinefun))
         result[branch==cur.branch,j] <- lpcsl[[cur.branch+1]]$splinefun[[j]](cur.or.pi)  # here cur.op.pi instead of or.pi
      }
    result
  }




lpc.fit.spline <- function(lpcsl,  num.knots=100, add.spline=FALSE, lty=1, col=3){
    nbranch <- length(lpcsl)-1   
    # First of all evaluate the spline at each knot
    knots.pi <- list()
    knots.coords <- list()
    for (r in 0:nbranch) {
      #knots.pi[[r+1]] <- lpcsl[[r+1]]$range[1] + 0:(num.knots-1)/num.knots * (lpcsl[[r+1]]$range[2]-lpcsl[[r+1]]$range[1])
      #print(lpcsl[[r+1]]$range)
      knots.pi[[r+1]] <- lpcsl[[r+1]]$range[1] + 0:(num.knots-1)/(num.knots-1) * (lpcsl[[r+1]]$range[2]-lpcsl[[r+1]]$range[1])
      coords <- matrix(ncol=num.knots,nrow=length(lpcsl[[r+1]]$splinefun))
      for (j in 1:length(lpcsl[[r+1]]$splinefun))
        coords[j,] <- lpcsl[[r+1]]$splinefun[[j]](knots.pi[[r+1]])
      knots.coords[[r+1]] <- coords
    }
    if (add.spline==TRUE){ 
      for (r in 0:nbranch){
         lines(knots.coords[[r+1]][1,],knots.coords[[r+1]][2,] ,lty=lty, col=col)
       }
    }

   list(knots.coords=knots.coords, knots.pi=knots.pi)
}




# function that computes the true length of the interpolated spline
# arguments are
# lpcsl : object created by lpc.splinefun
# or.pi : the original projection indices (i.e. on the linear interpolant)
# branch : the id of the branch the points are supposedly on

lpc.curve.length <- function(lpcsl,or.pi,branch=0,total.subdivisions=1e4,min.subdivisions=100) {
    gradnorm <- function(tt,sps) {
      sum <- numeric(length(tt))
      for (j in 1:length(sps))
        sum <- sum + sps[[j]](tt,deriv=1)^2
      sqrt(sum)
    }
    if (length(branch)==1)
      branch <- rep(branch,length(or.pi))
    if (length(or.pi)!=length(branch))
      stop("Length mismatch between or.pi and branch")
    result <- numeric(length(or.pi))
    branches <- unique(branch)
    for (cur.branch in branches) {
      cur.or.pi <- or.pi[branch==cur.branch]
      length.between.points <- numeric(length(cur.or.pi))
      cur.or.pi.order <- order(cur.or.pi)
      cur.or.pi <- cur.or.pi[cur.or.pi.order]
      #print(cur.or.pi)
      from <- c(min(cur.or.pi),cur.or.pi)
      for (i in 1:length(cur.or.pi.order)) {        
        subdivisions <- max(min.subdivisions,ceiling(total.subdivisions*(cur.or.pi[i]-from[i])/(lpcsl[[cur.branch+1]]$range[2]-lpcsl[[cur.branch+1]]$range[1])))
        length.between.points[i] <- integrate(gradnorm,lower=from[i],upper=cur.or.pi[i],subdivisions=subdivisions,sps=lpcsl[[cur.branch+1]]$splinefun)$value
      }
      total.length <- cumsum(length.between.points)
      #print(total.length)
      result[branch==cur.branch][cur.or.pi.order] <- total.length+min(cur.or.pi)
    }
    result
  } 



# This function computes the projections of new data points onto the curve, VERY SLOW!


lpc.project.spline <- function(lpcsl, newdata, num.knots=100, optimize=TRUE, add.proj=FALSE, lty=2, col=4) {
    squared.distance <- function(t,coords,sps) {
      dist <- 0
      for (i in 1:length(sps))
        dist <- dist + (coords[i]-sps[[i]](t))^2
      dist
    }
    nbranch <- length(lpcsl)-1
      
    # First of all evaluate the spline at each knot
    # knots.pi <- list()
    # knots.coords <- list()
    # for (r in 0:nbranch) {
    #  knots.pi[[r+1]] <- lpcsl[[r+1]]$range[1] + 0:(num.knots-1)/num.knots * (lpcsl[[r+1]]$range[2]-lpcsl[[r+1]]$range[1])
    #  coords <- matrix(ncol=num.knots,nrow=length(lpcsl[[r+1]]$splinefun))
    #  for (j in 1:length(lpcsl[[r+1]]$splinefun))
    #    coords[j,] <- lpcsl[[r+1]]$splinefun[[j]](knots.pi[[r+1]])
    #  knots.coords[[r+1]] <- coords
    # }

    fit.spline <- lpc.fit.spline(lpcsl, num.knots=num.knots)
    knots.coords  <- fit.spline[[1]]
    knots.pi      <- fit.spline[[2]] 
    newdata <- as.matrix(newdata)
   
  
    closest.branch <- numeric(nrow(newdata))
    closest.or.pi <- numeric(nrow(newdata))
    closest.idx <- numeric(nrow(newdata))
    closest.dist <- numeric(nrow(newdata))
 
     for (i in 1:nrow(newdata)) {
      sqdist <- Inf
      for (r in 0:nbranch) {
       
        cur.sqdist <- apply((knots.coords[[r+1]]-newdata[i,])^2,2,sum)
        if (min(cur.sqdist)<sqdist) {
          closest.branch[i] <- r
          closest.idx[i] <- which.min(cur.sqdist)
          closest.or.pi[i] <- knots.pi[[r+1]][closest.idx[i]]
          sqdist <- min(cur.sqdist)
         }
      }
      if (optimize) {
        from <- knots.pi[[closest.branch[i]+1]][max(1,closest.idx[i]-1)]
        to <- knots.pi[[closest.branch[i]+1]][min(num.knots,closest.idx[i]+1)]
       
        opt <- optimize(squared.distance,lower=from,upper=to,coords=newdata[i,],sps=lpcsl[[closest.branch[i]+1]]$splinefun)
       
        closest.or.pi[i] <- opt$minimum
        closest.dist[i] <- sqrt(opt$objective)
      } else {
        closest.dist[i] <- sqrt(squared.distance(closest.or.pi[i],coords=newdata[i,],sps=lpcsl[[closest.branch[i]+1]]$splinefun))
      }
    }
    closest.pi <- lpc.curve.length(lpcsl,closest.or.pi,closest.branch)
    closest.coords <- lpc.spline.eval(lpcsl, closest.or.pi, closest.branch)
   
    
    if (add.proj==TRUE){
      for (i in 1:nrow(newdata)){
         lines(rbind(newdata[i,],closest.coords[i,]),lty=lty,col=col) # here the $proj had to be removed
       }
    }

    result <- list(closest.branch=closest.branch,closest.pi=closest.pi,closest.or.pi=closest.or.pi,closest.coords=closest.coords,closest.dist=closest.dist)
    result
  }






